@@ -34,6 +34,7 @@
 #include <syslog.h>
 #include <errno.h>
 #include <time.h>
+#include <ctype.h>
 #include <X11/Xlib.h>
 
 #include <vicious.h>
@@ -45,15 +46,17 @@
 #include "display.h"
 #include "auth.h"
 #include "slave.h"
+#include "getvt.h"
 
 #define SERVER_WAIT_ALARM 10
 
 
 /* Local prototypes */
-static void gdm_server_spawn (GdmDisplay *d);
+static void gdm_server_spawn (GdmDisplay *d, const char *vtarg);
 static void gdm_server_usr1_handler (gint);
 static void gdm_server_alarm_handler (gint);
 static void gdm_server_child_handler (gint);
+static char * get_font_path (const char *display);
 
 /* Configuration options */
 extern gchar *GdmDisplayInit;
@@ -85,6 +88,19 @@ gdm_server_wipe_cookies (GdmDisplay *disp)
 	disp->authfile_gdm = NULL;
 }
 
+/* ignore handlers */
+static int
+ignore_xerror_handler (Display *disp, XErrorEvent *evt)
+{
+	return 0;
+}
+
+static int
+exit_xioerror_handler (Display *disp)
+{
+	_exit (1);
+}
+
 /**
  * gdm_server_reinit:
  * @disp: Pointer to a GdmDisplay structure
@@ -96,6 +112,8 @@ gdm_server_wipe_cookies (GdmDisplay *disp)
 void
 gdm_server_reinit (GdmDisplay *disp)
 {
+	pid_t pid;
+
 	if (disp == NULL)
 		return;
 
@@ -110,7 +128,65 @@ gdm_server_reinit (GdmDisplay *disp)
 
 	gdm_debug ("gdm_server_reinit: Server for %s is about to be reinitialized!", disp->name);
 
-	kill (disp->servpid, SIGHUP);
+	/* This hack below is beautiful BTW */
+
+	gdm_sigchld_block_push ();
+
+	pid = gdm_fork_extra ();
+	if (pid < 1) {
+		/* So much work just because we can't fork */
+		if (disp->servpid > 0)
+			kill (disp->servpid, SIGHUP);
+
+		gdm_sigchld_block_push ();
+
+		/* HACK! the Xserver can't really tell us when it got the hup signal,
+		 * so we are really stuck just going to sleep for a bit hoping that
+		 * the kernel will tell the X server and that will run, else we will
+		 * get whacked ourselves after we open the connection and we'll think
+		 * it's an X screwup, which is really OK to happen and will just
+		 * restart the Xserver, it's just more nasty.  Oh how fun */
+		sleep (1);
+	} else if (pid == 0) {
+		Display *dsp;
+
+		closelog ();
+
+		/* close things */
+		gdm_close_all_descriptors (0 /* from */, -1 /* except */);
+
+		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
+
+		XSetErrorHandler (ignore_xerror_handler);
+		XSetIOErrorHandler (exit_xioerror_handler);
+	       
+		gnome_setenv ("XAUTHORITY", d->authfile, TRUE);
+		dsp = XOpenDisplay (d->name);
+
+		/* Now whack the server with a SIGHUP */
+		if (disp->servpid > 0)
+			kill (disp->servpid, SIGHUP);
+
+		if (dsp == NULL) {
+			/* HACK, see note above */
+			sleep (1);
+			_exit (0);
+		} else {
+			/* Wait for an xioerror */
+			for (;;) {
+				XEvent event;
+				XNextEvent (dsp, &event);
+			}
+		}
+		/* anality */
+		_exit (0);
+	} else if (pid > 0) {
+		gdm_sigchld_block_pop ();
+
+		gdm_wait_for_extra (NULL);
+	}
 }
 
 /**
@@ -139,7 +215,7 @@ gdm_server_stop (GdmDisplay *disp)
 
     disp->servstat = SERVER_DEAD;
 
-    if (disp->servpid != 0) {
+    if (disp->servpid > 0) {
 	    pid_t servpid;
 
 	    gdm_debug ("gdm_server_stop: Killing server pid %d",
@@ -161,6 +237,8 @@ gdm_server_stop (GdmDisplay *disp)
     }
 
     gdm_server_wipe_cookies (disp);
+
+    gdm_slave_whack_temp_auth_file ();
 }
 
 static gboolean
@@ -257,7 +335,6 @@ display_busy (GdmDisplay *disp)
 	return FALSE;
 }
 
-#ifdef __linux__
 static int
 display_vt (GdmDisplay *disp)
 {
@@ -284,7 +361,38 @@ display_vt (GdmDisplay *disp)
 	fclose (fp);
 	return -1;
 }
-#endif
+
+static void
+check_child_status (void)
+{
+	int status;
+	pid_t pid;
+
+	while ((pid = waitpid (-1, &status, WNOHANG)) > 0) {
+		gdm_debug ("check_child_status: %d died", pid);
+
+		if (WIFEXITED (status))
+			gdm_debug ("check_child_status: %d returned %d",
+				   (int)pid, (int)WEXITSTATUS (status));
+		if (WIFSIGNALED (status))
+			gdm_debug ("check_child_status: %d died of %d",
+				   (int)pid, (int)WTERMSIG (status));
+
+		if (pid == d->servpid) {
+			gdm_debug ("check_child_status: Got SIGCHLD from server, "
+				   "server abort");
+
+			d->servstat = SERVER_ABORT;	/* Server died unexpectedly */
+			d->servpid = 0;
+
+			server_signal_notified = TRUE;
+		} else if (pid == extra_process) {
+			/* an extra process died, yay! */
+			extra_process = 0;
+			extra_status = status;
+		}
+	}
+}
 
 /**
  * gdm_server_start:
@@ -301,6 +409,8 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
     struct sigaction old_usr1, old_chld, old_alrm;
     sigset_t mask, oldmask;
     int flexi_disp = 20;
+    char *vtarg = NULL;
+    int vtfd = -1, vt = -1;
     
     if (disp == NULL)
 	    return FALSE;
@@ -310,14 +420,12 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
     /* if an X server exists, wipe it */
     gdm_server_stop (d);
 
-    /* On linux first clear the VT number */
-#ifdef __linux__
+    /* First clear the VT number */
     if (d->type == TYPE_LOCAL ||
 	d->type == TYPE_FLEXI) {
 	    d->vt = -1;
 	    gdm_slave_send_num (GDM_SOP_VT_NUM, -1);
     }
-#endif
 
     if (SERVER_IS_FLEXI (d) ||
 	treat_as_flexi) {
@@ -409,8 +517,13 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 
     d->servstat = SERVER_DEAD;
 
+    if (d->type == TYPE_LOCAL ||
+	d->type == TYPE_FLEXI) {
+	    vtarg = gdm_get_empty_vt_argument (&vtfd, &vt);
+    }
+
     /* fork X server process */
-    gdm_server_spawn (d);
+    gdm_server_spawn (d, vtarg);
 
     /* we can now use d->handled since that's set up above */
 
@@ -423,6 +536,8 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 		     * fortunately there is no such case yet and probably
 		     * never will, but just for code anality's sake */
 		    sleep (5);
+		    /* In case we got a SIGCHLD */
+		    check_child_status ();
 	    } else if (d->server_uid != 0) {
 		    int i;
 
@@ -432,6 +547,9 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 		     * use USR1 of course, so try openning the display 
 		     * as a test, but the */
 
+		    /* In case we got a SIGCHLD */
+		    check_child_status ();
+
 		    gnome_setenv ("XAUTHORITY", d->authfile, TRUE);
 		    for (i = 0;
 			 d->dsp == NULL &&
@@ -443,6 +561,9 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 				    sleep (1);
 			    else
 				    d->servstat = SERVER_RUNNING;
+
+			    /* In case we got a SIGCHLD */
+			    check_child_status ();
 		    }
 		    gnome_unsetenv ("XAUTHORITY");
 		    if (d->dsp == NULL &&
@@ -460,17 +581,27 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 
 		    do {
 			    select (server_signal_pipe[0]+1, &rfds, NULL, NULL, NULL);
+			    /* In case we got a SIGCHLD */
+			    check_child_status ();
 		    } while ( ! server_signal_notified);
 
 		    gdm_debug ("gdm_server_start: After mainloop waiting for server");
 	    }
     }
 
+    /* In case we got a SIGCHLD */
+    check_child_status ();
+
     /* Unset alarm */
     if (d->server_uid == 0) {
 	    alarm (0);
     }
 
+    /* If we were holding a vt open for the server, close it now as it has
+     * already taken the bait. */
+    if (vtfd > 0)
+	    close (vtfd);
+
     switch (d->servstat) {
 
     case SERVER_TIMEOUT:
@@ -496,14 +627,16 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 
 	    if (SERVER_IS_FLEXI (d))
 		    gdm_slave_send_num (GDM_SOP_FLEXI_OK, 0 /* bogus */);
-#ifdef __linux__
 	    if (d->type == TYPE_LOCAL ||
 		d->type == TYPE_FLEXI) {
-		    d->vt = display_vt (d);
+		    if (vt >= 0)
+			    d->vt = vt;
+
+		    if (d->vt < 0)
+			    d->vt = display_vt (d);
 		    if (d->vt >= 0)
 			    gdm_slave_send_num (GDM_SOP_VT_NUM, d->vt);
 	    }
-#endif
 
 	    return TRUE;
     default:
@@ -514,6 +647,8 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
     if (d->servpid > 0) {
 	    pid_t pid;
 
+	    d->dsp = NULL;
+
 	    gdm_sigchld_block_push ();
 	    pid = d->servpid;
 	    d->servpid = 0;
@@ -528,6 +663,9 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 
     sigprocmask (SIG_SETMASK, &oldmask, NULL);
 
+    /* In case we got a SIGCHLD */
+    check_child_status ();
+
     /* restore default handlers */
     sigaction (SIGUSR1, &old_usr1, NULL);
     sigaction (SIGCHLD, &old_chld, NULL);
@@ -594,7 +732,6 @@ gdm_server_start (GdmDisplay *disp, gboolean treat_as_flexi,
 void
 gdm_server_checklog (GdmDisplay *disp)
 {
-#ifdef __linux__
 	if (d->vt < 0 &&
 	    (d->type == TYPE_LOCAL ||
 	     d->type == TYPE_FLEXI)) {
@@ -602,7 +739,6 @@ gdm_server_checklog (GdmDisplay *disp)
 		if (d->vt >= 0)
 			gdm_slave_send_num (GDM_SOP_VT_NUM, d->vt);
 	}
-#endif
 }
 
 static void
@@ -629,14 +765,17 @@ rotate_logs (const char *dname)
 	g_free (fname);
 }
 
+
 char **
 gdm_server_resolve_command_line (GdmDisplay *disp,
-				 gboolean resolve_handled)
+				 gboolean resolve_handled,
+				 const char *vtarg)
 {
 	char *bin;
 	char **argv;
 	int len;
 	int i;
+	gboolean gotvtarg = FALSE;
 
 	bin = ve_first_word (disp->command);
 	if (bin == NULL) {
@@ -671,23 +810,36 @@ gdm_server_resolve_command_line (GdmDisplay *disp,
 		argv = ve_split (disp->command);
 	}
 
-	for (len = 0; argv != NULL && argv[len] != NULL; len++)
-		;
+	for (len = 0; argv != NULL && argv[len] != NULL; len++) {
+		char *arg = argv[len];
+		/* HACK! Not to add vt argument to servers that already force
+		 * allocation.  Mostly for backwards compat only */
+		if (strncmp (arg, "vt", 2) == 0 &&
+		    isdigit (arg[2]) &&
+		    (arg[3] == '\0' ||
+		     (isdigit (arg[3]) && arg[4] == '\0')))
+			gotvtarg = TRUE;
+	}
 
-	argv = g_renew (char *, argv, len + 4);
+	argv = g_renew (char *, argv, len + 5);
 	for (i = len - 1; i >= 1; i--) {
 		argv[i+1] = argv[i];
 	}
 	/* server number is the FIRST argument, before any others */
 	argv[1] = g_strdup (disp->name);
+	len++;
 	if (disp->authfile != NULL) {
-		argv[len+1] = g_strdup ("-auth");
-		argv[len+2] = g_strdup (disp->authfile);
-		argv[len+3] = NULL;
-	} else {
-		argv[len+1] = NULL;
+		argv[len++] = g_strdup ("-auth");
+		argv[len++] = g_strdup (disp->authfile);
 	}
 
+	if (vtarg != NULL &&
+	    ! gotvtarg) {
+		argv[len++] = g_strdup (vtarg);
+	}
+
+	argv[len++] = NULL;
+
 	g_free (bin);
 
 	return argv;
@@ -704,7 +856,7 @@ gdm_server_resolve_command_line (GdmDisplay *disp,
  */
 
 static void 
-gdm_server_spawn (GdmDisplay *d)
+gdm_server_spawn (GdmDisplay *d, const char *vtarg)
 {
     struct sigaction ign_signal, dfl_signal;
     sigset_t mask;
@@ -732,7 +884,8 @@ gdm_server_spawn (GdmDisplay *d)
 
     /* Figure out the server command */
     argv = gdm_server_resolve_command_line (d,
-					    TRUE /* resolve handled */);
+					    TRUE /* resolve handled */,
+					    vtarg);
 
     command = g_strjoinv (" ", argv);
 
@@ -760,9 +913,9 @@ gdm_server_spawn (GdmDisplay *d)
 
 	/* No error checking here - if it's messed the best response
          * is to ignore & try to continue */
-	open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-	open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-	open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+	gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+	gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+	gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 
 	openlog ("gdm", LOG_PID, LOG_DAEMON);
 
@@ -819,11 +972,26 @@ gdm_server_spawn (GdmDisplay *d)
 	sigprocmask (SIG_SETMASK, &mask, NULL);
 
 	if (d->type == TYPE_FLEXI_XNEST) {
+		char *font_path = NULL;
 		gnome_setenv ("DISPLAY", d->xnest_disp, TRUE);
 		if (d->xnest_auth_file != NULL)
 			gnome_setenv ("XAUTHORITY", d->xnest_auth_file, TRUE);
 		else
 			gnome_unsetenv ("XAUTHORITY");
+
+		/* Add -fp with the current font path, but only if not
+		 * already among the arguments */
+		if (strstr (command, "-fp") == NULL)
+			font_path = get_font_path (d->xnest_disp);
+		if (font_path != NULL) {
+			int argc = ve_vector_len (argv);
+			argv = g_renew (char *, argv, argc + 3);
+			argv[argc++] = "-fp";
+			argv[argc++] = font_path;
+			argv[argc++] = NULL;
+			command = g_strconcat (command, " -fp ",
+					       font_path, NULL);
+		}
 	}
 
 	if (argv[0] == NULL) {
@@ -966,39 +1134,12 @@ gdm_server_alarm_handler (gint signal)
 static void 
 gdm_server_child_handler (int signal)
 {
-	int status;
-	pid_t pid;
-
 	gdm_in_signal++;
 
 	gdm_debug ("gdm_server_child_handler: Got SIGCHLD");
 
-	while ((pid = waitpid (-1, &status, WNOHANG)) > 0) {
-		gdm_debug ("gdm_server_child_handler: %d died", pid);
-
-		if (WIFEXITED (status))
-			gdm_debug ("gdm_server_child_handler: %d returned %d",
-				   (int)pid, (int)WEXITSTATUS (status));
-		if (WIFSIGNALED (status))
-			gdm_debug ("gdm_server_child_handler: %d died of %d",
-				   (int)pid, (int)WTERMSIG (status));
-
-		if (pid == d->servpid) {
-			gdm_debug ("gdm_server_child_handler: Got SIGCHLD from server, "
-				   "server abort");
-
-			d->servstat = SERVER_ABORT;	/* Server died unexpectedly */
-			d->servpid = 0;
-
-			server_signal_notified = TRUE;
-			/* this will quit the select */
-			write (server_signal_pipe[1], "Yay!", 4);
-		} else if (pid == extra_process) {
-			/* an extra process died, yay! */
-			extra_process = -1;
-			extra_status = status;
-		}
-	}
+	/* this will quit the select */
+	write (server_signal_pipe[1], "Yay!", 4);
 
 	gdm_in_signal--;
 }
@@ -1039,7 +1180,6 @@ gdm_server_alloc (gint id, const gchar *command)
     d->dispnum = id;
     d->servpid = 0;
     d->servstat = SERVER_DEAD;
-    d->sessionid = 0;
     d->sesspid = 0;
     d->slavepid = 0;
     d->type = TYPE_LOCAL;
@@ -1052,9 +1192,7 @@ gdm_server_alloc (gint id, const gchar *command)
 
     d->handled = TRUE;
 
-#ifdef __linux__
     d->vt = -1;
-#endif
 
     d->x_servers_order = -1;
 
@@ -1072,13 +1210,6 @@ gdm_server_alloc (gint id, const gchar *command)
     return d;
 }
 
-/* ignore handlers */
-static int
-ignore_xerror_handler (Display *disp, XErrorEvent *evt)
-{
-	return 0;
-}
-
 void
 gdm_server_whack_clients (GdmDisplay *disp)
 {
@@ -1120,5 +1251,37 @@ gdm_server_whack_clients (GdmDisplay *disp)
 	XSetErrorHandler (old_xerror_handler);
 }
 
+static char *
+get_font_path (const char *display)
+{
+	Display *disp;
+	char **font_path;
+	int n_fonts;
+	int i;
+	GString *gs;
+
+	disp = XOpenDisplay (display);
+	if (disp == NULL)
+		return NULL;
+
+	font_path = XGetFontPath (disp, &n_fonts);
+	if (font_path == NULL) {
+		XCloseDisplay (disp);
+		return NULL;
+	}
+
+	gs = g_string_new (NULL);
+	for (i = 0; i < n_fonts; i++) {
+		if (i != 0)
+			g_string_append_c (gs, ',');
+		g_string_append (gs, font_path[i]);
+	}
+
+	XFreeFontPath (font_path);
+
+	XCloseDisplay (disp);
+
+	return g_string_free (gs, FALSE);
+}
 
 /* EOF */
